use nalgebra::Matrix2x3;
use std::f64::consts::PI as PI;
use sycamore::{prelude::*, rt::{JsCast, JsValue}};

mod engine;

fn main() {
  // set up a config option that forwards panic messages to `console.error`
  #[cfg(feature = "console_error_panic_hook")]
  console_error_panic_hook::set_once();
  
  sycamore::render(|| {
    let data = [-1.0, 0.0, 0.0, -1.0, 1.0, 0.0].map(|n| create_signal(n));
    let display = create_node_ref();
    
    on_mount(move || {
      let canvas = display
        .get::<DomNode>()
        .unchecked_into::<web_sys::HtmlCanvasElement>();
      let ctx = canvas
        .get_context("2d")
        .unwrap()
        .unwrap()
        .dyn_into::<web_sys::CanvasRenderingContext2d>()
        .unwrap();
      
      create_effect(move || {
        // center and normalize the coordinate system
        let width = canvas.width() as f64;
        let height = canvas.height() as f64;
        ctx.set_transform(1.0, 0.0, 0.0, -1.0, 0.5*width, 0.5*height).unwrap();
        
        // clear the previous frame
        ctx.clear_rect(-0.5*width, -0.5*width, width, height);
        
        // find the resolution
        const R_DISP: f64 = 5.0;
        let res = width / (2.0*R_DISP);
        
        // set colors
        let highlight_style = JsValue::from("white");
        let grid_style = JsValue::from("#404040");
        let point_fill_styles = ["#ba5d09", "#0e8a06", "#8951fb"];
        let point_stroke_styles = ["#f89142", "#58c145", "#c396fc"];
        
        // draw the grid
        let r_grid = (R_DISP - 0.01).floor() as i32;
        let edge_scr = res * R_DISP;
        ctx.set_stroke_style(&grid_style);
        for t in -r_grid ..= r_grid {
          let t_scr = res * (t as f64);
          
          // draw horizontal grid line
          ctx.begin_path();
          ctx.move_to(-edge_scr, t_scr);
          ctx.line_to(edge_scr, t_scr);
          ctx.stroke();
          
          // draw vertical grid line
          ctx.begin_path();
          ctx.move_to(t_scr, -edge_scr);
          ctx.line_to(t_scr, edge_scr);
          ctx.stroke();
        }
        
        // find and draw the circle through the given points
        let data_vals = data.map(|sig| sig.get()).to_vec();
        let points = Matrix2x3::from_vec(data_vals);
        if let Some(circ) = engine::circ_thru(points) {
          ctx.begin_path();
          ctx.set_stroke_style(&highlight_style);
          ctx.arc(
            res * circ.center_x,
            res * circ.center_y,
            res * circ.radius,
            0.0, 2.0*PI
          ).unwrap();
          ctx.stroke();
        }
        
        // draw the data points
        for n in 0..3 {
          ctx.begin_path();
          ctx.set_fill_style(&JsValue::from(point_fill_styles[n]));
          ctx.set_stroke_style(&JsValue::from(point_stroke_styles[n]));
          let ind_x = 2*n;
          let ind_y = ind_x + 1;
          ctx.arc(
            res * data[ind_x].get(),
            res * data[ind_y].get(),
            3.0,
            0.0, 2.0*PI
          ).unwrap();
          ctx.fill();
          ctx.stroke();
        }
      });
    });
    
    view! {
      canvas(ref=display, width="600", height="600")
      div(id="data-panel") {
        div { "x" }
        div { "y" }
        input(type="number", class="point-1", bind:valueAsNumber=data[0])
        input(type="number", class="point-1", bind:valueAsNumber=data[1])
        input(type="number", class="point-2", bind:valueAsNumber=data[2])
        input(type="number", class="point-2", bind:valueAsNumber=data[3])
        input(type="number", class="point-3", bind:valueAsNumber=data[4])
        input(type="number", class="point-3", bind:valueAsNumber=data[5])
      }
    }
  });
}